from typing import Iterator

import requests
from requests.structures import CaseInsensitiveDict

from aistore.sdk.const import DEFAULT_CHUNK_SIZE
from aistore.sdk.object_attributes import ObjectAttributes


class ObjectReader:
    """
    Represents the data returned by the API when getting an object, including access to the content stream and object
    attributes
    """

    def __init__(
        self,
        response_headers: CaseInsensitiveDict,
        stream: requests.Response,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
    ):
        self._chunk_size = chunk_size
        self._stream = stream
        self._attributes = ObjectAttributes(response_headers)

    @property
    def attributes(self) -> ObjectAttributes:
        """
        Object metadata attributes

        Returns:
            Object attributes parsed from the headers returned by AIS
        """
        return self._attributes

    def read_all(self) -> bytes:
        """
        Read all byte data from the object content stream.
        This uses a bytes cast which makes it slightly slower and requires all object content to fit in memory at once
        Returns:
            Object content as bytes

        """
        obj_arr = bytearray()
        for chunk in self:
            obj_arr.extend(chunk)
        return bytes(obj_arr)

    def raw(self) -> bytes:
        """
        Returns: Raw byte stream of object content
        """
        return self._stream.raw

    def __iter__(self) -> Iterator[bytes]:
        """
        Creates a generator to read the stream content in chunks
        Returns:
            An iterator with access to the next chunk of bytes
        """
        try:
            for chunk in self._stream.iter_content(chunk_size=self._chunk_size):
                yield chunk
        finally:
            self._stream.close()



"""
This module is a temporary patch to support the torchdata integration for AIStore SDK versions > 1.04.
Torchdata expects the sdk structure to have aistore.client.Client, but Client now exists in aistore.sdk.client
"""

# pylint: disable=unused-variable,unused-import
from aistore.sdk.client import Client



#
# Copyright (c) 2023, NVIDIA CORPORATION. All rights reserved.
#
from typing import Iterator

from aistore.sdk import utils
from aistore.sdk.errors import InvalidObjectRangeIndex
from aistore.sdk.multiobj.object_collection import ObjectCollection


# pylint: disable=unused-variable,too-few-public-methods
class ObjectRange(ObjectCollection):
    """
    Class representing a range of object names

    Args:
        prefix (str): Prefix contained in all names of objects
        min_index (int): Starting index in the name of objects
        max_index (int): Last index in the name of all objects
        pad_width (int, optional): Left-pad indices with zeros up to the width provided, e.g. pad_width = 3 will
            transform 1 to 001
        step (int, optional): Size of iterator steps between each item
        suffix (str, optional): Suffix at the end of all object names
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        prefix: str,
        min_index: int = None,
        max_index: int = None,
        pad_width: int = 0,
        step: int = 1,
        suffix: str = "",
    ):
        self._prefix = prefix
        self._step = step
        self._suffix = suffix
        self._validate_indices(min_index, max_index, pad_width)
        min_set = isinstance(min_index, int)
        if self._suffix and not min_set:
            raise ValueError("Suffix cannot be used without indices")
        self._min_index = str(min_index).zfill(pad_width) if min_set else None
        self._max_index = (
            str(max_index).zfill(pad_width) if isinstance(min_index, int) else None
        )

    @staticmethod
    def _validate_indices(min_index, max_index, pad_width):
        """
        Validate the indices passed to create a range: min_index < max_index and pad_width (if set) can fit the indices
            provided.

        Raises:
            InvalidObjectRangeIndex: If the indices passed to the range are not valid
        """
        indices = [isinstance(min_index, int), isinstance(max_index, int)]
        if not all(indices):
            if any(indices):
                raise InvalidObjectRangeIndex(
                    "Provide both min_index and max_index or neither"
                )
            return

        if min_index >= max_index:
            raise InvalidObjectRangeIndex(f"{min_index} must be less than {max_index}")
        if pad_width != 0 and len(str(min_index)) > pad_width:
            raise InvalidObjectRangeIndex(
                f"Number of digits in min index {min_index} must not be greater than pad width {pad_width}"
            )

    def _indices_set(self) -> bool:
        if all([self._min_index, self._max_index]):
            return True
        return False

    def __str__(self) -> str:
        if self._indices_set():
            return f"{self._prefix}{{{self._min_index}..{self._max_index}..{self._step}}}{self._suffix}"
        return f"{self._prefix}"

    def __iter__(self) -> Iterator[str]:
        if not self._indices_set():
            raise RuntimeError("Cannot iterate over object range with no indices")
        return utils.expand_braces(str(self))

    def get_value(self):
        return {"template": str(self)}



from aistore.sdk.multiobj.object_collection import ObjectCollection
from aistore.sdk.multiobj.object_template import ObjectTemplate
from aistore.sdk.multiobj.object_range import ObjectRange
from aistore.sdk.multiobj.object_names import ObjectNames
from aistore.sdk.multiobj.object_group import ObjectGroup



#
# Copyright (c) 2023, NVIDIA CORPORATION. All rights reserved.
#

from typing import Dict, Iterator

from aistore.sdk import utils
from aistore.sdk.multiobj.object_collection import ObjectCollection


class ObjectTemplate(ObjectCollection):
    """
    A collection of object names specified by a template in the bash brace expansion format

    Args:
         template (str): A string template that defines the names of objects to include in the collection

    """

    def __init__(self, template: str):
        self._template = template

    def __iter__(self) -> Iterator[str]:
        return utils.expand_braces(self._template)

    def get_value(self) -> Dict[str, str]:
        return {"template": self._template}
